@@ -1,7 +1,7 @@
 /*
  * bufio.c - Buffered, selectable I/O on pipes.
  *
- * Copyright (c) 1994, 1995 Eric Schenk.
+ * Copyright (c) 1994, 1995, 1996 Eric Schenk.
  * All rights reserved. Please see the file LICENSE which should be
  * distributed with this software for terms of use.
  */
@@ -10,9 +10,12 @@
 
 void pipe_init(int fd,PIPE *pipe)
 {
+    char buf[2];
     pipe->fd = fd;
     pipe->count = 0;
     fcntl(fd,F_SETFL,fcntl(fd,F_GETFL)|O_NONBLOCK);
+    /* clear out any old garbage from the FIFO */
+    while (read(fd,buf,1) > 0);
 }
 
 /* Read from the file descriptor, and
