@@ -1,7 +1,7 @@
 /*
  * diald.c - Demand dialing daemon for ppp.
  *
- * Copyright (c) 1994, 1995 Eric Schenk.
+ * Copyright (c) 1994, 1995, 1996 Eric Schenk.
  * All rights reserved. Please see the file LICENSE which should be
  * distributed with this software for terms of use.
  *
@@ -30,21 +30,48 @@
 
 /* intialized variables. */
 int modem_fd = -1;		/* modem device fp (for proxy reads) */
-FILE *mon_fp = 0;		/* monitor outut fd */
+MONITORS *monitors = 0;		/* Monitor pipes */
 int modem_hup = 0;		/* have we seen a modem HUP? */
+int sockfd = -1;		/* controling socket */
 int delayed_quit = 0;		/* has the user requested a delayed quit? */
 int request_down = 0;		/* has the user requested link down? */
 int request_up = 0;		/* has the user requested link down? */
 int forced = 0;			/* has the user requested the link forced up? */
 int link_pid = 0;		/* current protocol control command pid */
 int dial_pid = 0;		/* current dial command pid */
+int running_pid = 0;		/* current system command pid */
+int running_status = 0;		/* status of last system command */
 int dial_status = 0;		/* status from last dial command */
 int state_timeout = -1;		/* state machine timeout counter */
 int proxy_iface = 0;		/* Interface for the proxy */
 int link_iface = -1;		/* Interface for the link */
+int force_dynamic = 0;		/* true if connect passed back an addr */
+int redial_rtimeout = -1;	/* initialized value */
+int dial_failures = 0;		/* count of dialing failures */
+int ppp_half_dead = 0;		/* is the ppp link half dead? */
 int terminate = 0;
 char *pidfile = 0;
 static PIPE fifo_pipe;
+int argc_save;
+char **argv_save;
+
+void do_config(void)
+{
+    init_vars();
+    flush_prules();
+    flush_vars();
+    flush_filters();
+    /* Get the default defs and config files first */
+    parse_options_file(DIALD_DEFS_FILE);
+    parse_options_file(DIALD_CONFIG_FILE);
+    /* Get the command line modifications */
+    parse_args(argc_save-1,argv_save+1);
+    /* Do validity checks on the setup */
+    check_setup();
+    /* FIXME: minor memory leak here */
+    orig_local_ip = strdup(local_ip);
+    orig_remote_ip = strdup(remote_ip);
+}
 
 void main(int argc, char *argv[])
 {
@@ -52,6 +79,9 @@ void main(int argc, char *argv[])
     struct timeval timeout;
     fd_set readfds;
 
+    argc_save = argc;
+    argv_save = argv;
+
     /* initialize system log interface */
     openlog("diald", LOG_PID | LOG_NDELAY | LOG_PERROR,  LOG_LOCAL2);
 
@@ -59,20 +89,19 @@ void main(int argc, char *argv[])
     /* If I get things into a device this should be an "open" */
     fwunit = ctl_firewall(IP_FW_OPEN,0);
 
-    /* Get the default defs and config files first */
     parse_init();
-    parse_options_file(DIALD_DEFS_FILE);
-    parse_options_file(DIALD_CONFIG_FILE);
-    /* Get the command line modifications */
-    parse_args(argc-1,argv+1);
-    /* Do validity checks on the setup */
-    check_setup();
-
-    orig_local_ip = strdup(local_ip);
-    orig_remote_ip = strdup(remote_ip);
+    do_config();
 
     become_daemon();
 
+    /* Get an internet socket for doing socket ioctls. */
+    sockfd = socket(AF_INET, SOCK_DGRAM, 0);
+    if (sockfd < 0)
+      {
+	syslog(LOG_ERR, "Couldn't create IP socket: %m");
+	die(1);
+      }
+
     open_fifo();
 
     if (debug&DEBUG_VERBOSE)
@@ -88,7 +117,7 @@ void main(int argc, char *argv[])
     idle_filter_proxy();
 
     if (debug&DEBUG_VERBOSE)
-	syslog(LOG_INFO,"Daild initial setup completed.");
+	syslog(LOG_INFO,"Diald initial setup completed.");
 
     /* main loop */
     timeout.tv_sec = PAUSETIME;
@@ -117,7 +146,7 @@ void main(int argc, char *argv[])
 	    timeout.tv_usec = 0;
 	    if (state_timeout > 0) state_timeout--;
 	    if (debug&DEBUG_TICK)
-	        syslog(LOG_DEBUG,"--- tick --- state %d block %d",state,blocked);
+	        syslog(LOG_DEBUG,"--- tick --- state %d block %d state_timeout %d",state,blocked,state_timeout);
 	    monitor_queue();
 	}
 	change_state();
@@ -152,6 +181,8 @@ void become_daemon()
     if ((fp = fopen(pidfile,"w")) != NULL) {
         fprintf(fp,"%d\n",getpid());
         fclose(fp);
+    } else {
+	syslog(LOG_ERR,"Unable to create run file %s: %m",pidfile);
     }
 }
 
@@ -168,7 +199,13 @@ void open_fifo()
 	    mknod(fifoname, S_IFIFO|0277, 0);
 	    chmod(fifoname, 0600);
 	}
-	if ((fifo_fd = open(fifoname, O_RDONLY | O_NDELAY)) >= 0) {
+	/* We need to open this RDWR to make select() work the
+         * way we want in kernels after 1.3.81. In particular
+	 * we don't want select() to return 1 whenever there
+	 * are no writers on the remote side of the command fifo.
+	 * This guarantees that there is always at least one writer...
+         */
+	if ((fifo_fd = open(fifoname, O_RDWR | O_NDELAY)) >= 0) {
 	    if (debug&DEBUG_VERBOSE)
 	   	 syslog(LOG_INFO,"Using fifo %s",fifoname);
 	} else {
@@ -185,21 +222,22 @@ void open_fifo()
 /*
  * Set up the signal handlers.
  */
+static sigset_t sig_mask;
 
 void signal_setup()
 {
-    sigset_t mask;
     struct sigaction sa;
     /* set up signal handlers */
 
-    sigemptyset(&mask);
-    sigaddset(&mask, SIGHUP);
-    sigaddset(&mask, SIGINT);
-    sigaddset(&mask, SIGTERM);
-    sigaddset(&mask, SIGUSR1);
-    sigaddset(&mask, SIGUSR2);
-    sigaddset(&mask, SIGCHLD);
-    sigaddset(&mask, SIGALRM);
+    sigemptyset(&sig_mask);
+    sigaddset(&sig_mask, SIGHUP);
+    sigaddset(&sig_mask, SIGINT);
+    sigaddset(&sig_mask, SIGTERM);
+    sigaddset(&sig_mask, SIGUSR1);
+    sigaddset(&sig_mask, SIGUSR2);
+    sigaddset(&sig_mask, SIGCHLD);
+    sigaddset(&sig_mask, SIGALRM);
+    sigaddset(&sig_mask, SIGPIPE);
 
 #define SIGNAL(s, handler)      { \
         sa.sa_handler = handler; \
@@ -209,8 +247,9 @@ void signal_setup()
         } \
     }
 
-    sa.sa_mask = mask;
+    sa.sa_mask = sig_mask;
     sa.sa_flags = 0;
+
     SIGNAL(SIGHUP, sig_hup);            /* Hangup: modem went down. */
     SIGNAL(SIGINT, sig_intr);           /* Interrupt: take demand dialer down */
     SIGNAL(SIGTERM, sig_term);          /* Terminate: user take link down */
@@ -218,6 +257,33 @@ void signal_setup()
     SIGNAL(SIGUSR2, print_filter_queue); /* dump the packet queue to the log */
     SIGNAL(SIGCHLD, sig_chld);		/* reap dead kids */
     SIGNAL(SIGALRM, alrm_timer);	/* Deal with a timer expired */
+    SIGNAL(SIGPIPE, SIG_IGN);
+}
+
+void block_signals()
+{
+    sigprocmask(SIG_BLOCK, &sig_mask, NULL);
+}
+
+void default_sigacts()
+{
+    struct sigaction sa;
+    sa.sa_mask = sig_mask;
+    sa.sa_flags = 0;
+
+    SIGNAL(SIGHUP, SIG_DFL);
+    SIGNAL(SIGINT, SIG_DFL);
+    SIGNAL(SIGTERM, SIG_DFL);
+    SIGNAL(SIGUSR1, SIG_DFL);
+    SIGNAL(SIGUSR2, SIG_DFL);
+    SIGNAL(SIGCHLD, SIG_DFL);
+    SIGNAL(SIGALRM, SIG_DFL);
+    SIGNAL(SIGPIPE, SIG_DFL);
+}
+
+void unblock_signals()
+{
+    sigprocmask(SIG_UNBLOCK, &sig_mask, NULL);
 }
 
 /*
@@ -261,13 +327,16 @@ void get_pty(int *mfd, int *sfd)
  *			  already finished and any lock files are in place.
  *			  When the connection should be killed we send a
  *			  SIGTERM to the given pid.
+ *	dynamic <lip> <rip> - pass back dynamic IP config info to diald.
+ *	message <txt>	- set the message text from the connect script.
  *	monitor file	- start a monitoring program.
+ *      reset		- reread the configuration information.
  */
 
 void fifo_read()
 {
     int i;
-    int pid, dev;
+    int pid, dev, j,k,l;
     char *buf, *tail;
 
     i = pipe_read(&fifo_pipe);
@@ -275,87 +344,114 @@ void fifo_read()
     if (i <= 0) return;	/* FIXME: swallow errors */
 
     while (i--) {
-       if (*tail == '\n') {
-           *tail = '\0';
-	   /* Ok, we've got a line, now we need to "parse" it. */
-	   if (strcmp(buf,"block") == 0) {
+        if (*tail == '\n') {
+            *tail = '\0';
+	    /* Ok, we've got a line, now we need to "parse" it. */
+	    if (strcmp(buf,"block") == 0) {
 		syslog(LOG_INFO, "FIFO: Block request received.");
 		blocked = 1;
-	   } else if (strcmp(buf,"unblock") == 0) {
+	    } else if (strcmp(buf,"unblock") == 0) {
 		syslog(LOG_INFO, "FIFO: Unblock request received.");
 		blocked = 0;
-	   } else if (strcmp(buf,"force") == 0) {
+	    } else if (strcmp(buf,"force") == 0) {
 		syslog(LOG_INFO, "FIFO: Force request received.");
 		forced = 1;
-	   } else if (strcmp(buf,"unforce") == 0) {
+	    } else if (strcmp(buf,"unforce") == 0) {
 		syslog(LOG_INFO, "FIFO: Unforce request received.");
 		forced = 0;
-	   } else if (strcmp(buf,"down") == 0) {
+	    } else if (strcmp(buf,"down") == 0) {
 		syslog(LOG_INFO, "FIFO: Link down request received.");
     		request_down = 1;
     		request_up = 0;
-	   } else if (strcmp(buf,"up") == 0) {
+	    } else if (strcmp(buf,"up") == 0) {
     		syslog(LOG_INFO, "FIFO: Link up request received.");
     		request_down = 0;
     		request_up = 1;
-	   } else if (strcmp(buf,"delayed-quit") == 0) {
+	    } else if (strcmp(buf,"delayed-quit") == 0) {
     		syslog(LOG_INFO, "FIFO. Delayed termination request received.");
     		delayed_quit = 1;
-	   } else if (strcmp(buf,"quit") == 0) {
+	    } else if (strcmp(buf,"quit") == 0) {
     		syslog(LOG_INFO, "FIFO. Termination request received.");
     		terminate = 1;
-	   } else if (strcmp(buf,"queue") == 0) {
+	    } else if (strcmp(buf,"reset") == 0) {
+    		syslog(LOG_INFO, "FIFO. Reset request received. Re-reading configuration.");
+		do_config();
+	    } else if (strcmp(buf,"queue") == 0) {
     		struct firewall_req req;
     		syslog(LOG_INFO,"FIFO. User requested dump of firewall queue.");
     		syslog(LOG_INFO,"--------------------------------------");
     		req.unit = fwunit;
     		ctl_firewall(IP_FW_PCONN,&req);
     		syslog(LOG_INFO,"--------------------------------------");
-	   } else if (sscanf(buf,"debug %d", &pid) == 1) {
+	    } else if (sscanf(buf,"debug %d", &pid) == 1) {
     		syslog(LOG_INFO,"FIFO. Changing debug flags to %d.",pid);
 		debug = pid;
-	   } else if (strncmp(buf,"monitor", 7) == 0) {
+	    } else if (sscanf(buf,"dynamic %n%*s%n %n",&j,&k,&l) == 1) {
+		buf[k] = 0;
+		if (inet_addr(buf+j) == (unsigned long)0xffffffff
+		||  inet_addr(buf+l) == (unsigned long)0xffffffff) {
+		    syslog(LOG_INFO,"FIFO: bad parameters '%s' and '%s' to dynamic command ignored", buf+j,buf+l);
+		} else {
+		    local_ip = strdup(buf+j);
+		    remote_ip = strdup(buf+l);
+		    force_dynamic = 1;
+		}
+	    } else if (strncmp(buf,"monitor", 7) == 0) {
     		struct stat sbuf;
+		int fd;
+		MONITORS *new;
 
-		if (mon_fp) { fclose(mon_fp); mon_fp = NULL; }
 		if (buf[7] != 0 && buf[7] == ' ') {
 		    syslog(LOG_INFO,"FIFO: monitor connection to %s requested",
 			    buf+8);
 		    if (stat(fifoname,&sbuf) < 0 || !sbuf.st_mode&S_IFIFO) {
 			syslog(LOG_INFO,"FIFO: %s not a pipe.",
 			    buf+8);
-		    } else if ((mon_fp = fopen(buf+8,"w"))==NULL) {
+		    } else if ((fd = open(buf+8,O_WRONLY))<0) {
 			syslog(LOG_INFO,"FIFO: could not open pipe %s: %m",
 			    buf+8);
-			mon_fp = 0;
 		    } else {
+			new = (MONITORS *)malloc(sizeof(MONITORS));
+			new->next = monitors;
+			new->fd = fd;
+			monitors = new;
 			output_state();
 		    }
 		} else {
-		    syslog(LOG_INFO,"FIFO: monitor connection closed");
+		    syslog(LOG_INFO,"FIFO: empty monitor request ignored");
 		}
-	   } else if (sscanf(buf,"connect %d %n", &pid, &dev) == 1) {
-               if (pid > 1) {
-		   if (state != STATE_DOWN || req_pid) {
-		       /* somebody else already has this diald, tell 'em */
-		       kill(pid, SIGTERM);
-		   } else {
-		       req_pid = pid;
-		       req_dev = (char *)malloc(tail-(buf+dev)+1);
-		       if (req_dev == 0) {
-			    req_pid = 0;
-		            syslog(LOG_ERR,"FIFO: no memory to store requested devce!");
-		       } else {
-		       	    strcpy(req_dev, buf+dev);
-		       	    request_down = 0;
-		            request_up = 1;
-		            syslog(LOG_INFO,"FIFO: link up requested on device %s", req_dev);
-		       }
-		   }
-	       }
-	   } else {
+	    } else if (strncmp(buf,"message ",8) == 0) {
+		/* pass a message from the connector on to the monitor */
+		if (monitors) {
+		    mon_write("MESSAGE\n",8);
+		    mon_write(buf+8,strlen(buf+8));
+		    mon_write("\n",1);
+		}
+            } else if (sscanf(buf,"connect %d %n", &pid, &dev) == 1) {
+                if (pid > 1) {
+		    if ((state != STATE_DOWN && state != STATE_CLOSE
+			&& !give_way)
+		    || state==STATE_UP || req_pid) {
+                        /* somebody else already has this diald, tell 'em */
+                        kill(pid, SIGTERM);
+			syslog(LOG_INFO,"FIFO: link up requested but denied");
+                    } else {
+                        req_pid = pid;
+                        req_dev = (char *)malloc(tail-(buf+dev)+1);
+                        if (req_dev == 0) {
+                            req_pid = 0;
+                            syslog(LOG_ERR,"FIFO: no memory to store requested devce!");
+                        } else {
+                            strcpy(req_dev, buf+dev);
+                            request_down = 0;
+                            request_up = 1;
+                            syslog(LOG_INFO,"FIFO: link up requested on device %s", req_dev);
+                        }
+                    }
+                }
+            } else {
 		syslog(LOG_ERR,"Unknown request '%s' made.", buf);
-	   }
+	    }
 	   buf = tail+1;
        }
        tail++;
@@ -374,23 +470,53 @@ void proxy_read()
 
     /* read the SLIP packet */
     len = recv_packet(buffer,4096);
-#ifndef UNSAFE_ROUTING
-    /* if we are doing unsafe routing, all counting is in the filter.
-     * otherwise we can see transmitted bytes directly at this spot.
+    if (!do_reroute) {
+	/* if we are doing unsafe routing, all counting is in the filter.
+	 * otherwise we can see transmitted bytes directly at this spot.
+	 */
+	txtotal += len;
+	itxtotal += len;
+	rxtotal -= len;	/* since it will double count on the snoop */
+	irxtotal -= len;
+    }
+
+    /* If we get here with the link up and fwdfd not -1,
+     * and we are rerouting, then it must be
+     * that the external interface has gone down without
+     * taking the link with it, and as a result our route
+     * to the external interface got lost. (This CAN legally
+     * happen with PPP). In this case we buffer the packet so
+     * we can retransmit it when the link comes back up.
+     * OR
+     * the kernel is retransmitting something through sl0, despite
+     * the existance of a route through another device...
      */
-    txtotal += len;
-    rxtotal -= len;	/* since it will double count on the snoop */
-#endif
-    if (link_iface != -1) {
+    if (link_iface != -1 && fwdfd != -1 && !(do_reroute && mode == MODE_PPP)) {
 	/* if the external iface is up then we can send it on */
-	struct sockaddr to;
+	struct SOCKADDR to;
+
+#ifdef HAS_SOCKADDR_PKT
+	to.spkt_family = AF_INET;
+	strcpy(to.spkt_device,snoop_dev);
+	to.spkt_protocol = htons(ETH_P_IP);
+#else
 	to.sa_family = AF_INET;
 	strcpy(to.sa_data,snoop_dev);
-	if (sendto(fwdfd,buffer,len,0,&to,sizeof(struct sockaddr)) < 0) {
+#endif
+	if (debug&DEBUG_VERBOSE)
+	    syslog(LOG_DEBUG,"Forwarding packet of length %d",len);
+	if (sendto(fwdfd,buffer,len,0,(struct sockaddr *)&to,sizeof(struct SOCKADDR)) < 0) {
 	    syslog(LOG_ERR,
 		"Error forwarding data packet to physical device: %m");
 	}
     } else {
+	/* Make sure we try to restore the link to working condition now... */
+	if (link_iface != -1 && fwdfd != -1 && do_reroute && mode == MODE_PPP) {
+	    /* Check if a route exists at this point through the ppp device. */
+	    /* If not then we must be half dead. */
+	    if (!ppp_route_exists())
+	        ppp_half_dead = 1;
+	}
 #ifdef BUFFER_PACKETS
 	/* The external iface is down, buffer the packet so we can
 	 * forward it when the iface comes up.
@@ -415,14 +541,16 @@ void die(int i)
 	/* We're killing without a care here. Uhggg. */
 	if (link_pid) kill(link_pid,SIGINT);
 	if (dial_pid) kill(dial_pid,SIGINT);
+	if (running_pid) kill(running_pid,SIGINT);
 	/* Wait up to 30 seconds for them to die */
         for (count = 0; (link_pid || dial_pid) && count < 30; count++)
 	    sleep(1);
 	/* If they aren't dead yet, kill them for sure */
 	if (link_pid) kill(link_pid,SIGKILL);
 	if (dial_pid) kill(dial_pid,SIGKILL);
+	if (running_pid) kill(running_pid,SIGKILL);
 	/* Give the system a second to send the signals */
-	if (link_pid || dial_pid) sleep(1);
+	if (link_pid || dial_pid || running_pid) sleep(1);
 	close_modem();
     	proxy_down();
 	unlink(pidfile);
@@ -473,11 +601,16 @@ void sig_chld(int sig)
     static int seq = 0;
     ++seq;
     while ((pid = waitpid(-1,&status,WNOHANG)) > 0) {
-	if (debug&DEBUG_VERBOSE)
-	    syslog( LOG_DEBUG, "SIGCHLD[%d]: pid %d %s", seq, pid,
-		    pid == link_pid ? "link" : pid == dial_pid ? "dial" : "other" );
+        if (debug&DEBUG_VERBOSE)
+	    syslog( LOG_DEBUG, "SIGCHLD[%d]: pid %d %s, status %d", seq, pid,
+		    pid == link_pid ? "link"
+		   	: pid == dial_pid ? "dial"
+			: pid == running_pid ? "system"
+			: "other",
+		    status);
 	if (pid == link_pid) link_pid = 0;
 	if (pid == dial_pid) { dial_status = status; dial_pid = 0; }
+	if (pid == running_pid) { running_status = status; running_pid = 0; }
 	if (pid > 0) {
 	    if (WIFSIGNALED(status)) {
 		syslog(LOG_WARNING, "child process %d terminated with signal %d",
@@ -498,3 +631,103 @@ void sig_term(int sig)
     syslog(LOG_INFO, "SIGTERM. Termination request received.");
     terminate = 1;
 }
+
+int report_system_result(int res,char *buf)
+{
+    if (res == -1)
+   	syslog(LOG_ERR,"System call failure on command '%s'",buf);
+    else if (!WIFEXITED(res))
+   	syslog(LOG_ERR,"Abnormal exit (status %d) on command '%s'",res,buf);
+    else if (WEXITSTATUS(res) != 0)
+	syslog(LOG_ERR,"Nonzero exit status (%d) on command '%s'",WEXITSTATUS(res),buf);
+    else
+	return 0;
+    return 1;
+}
+
+int system(const char *buf)
+{
+    int fd;
+
+    block_signals();
+    running_pid = fork();
+
+    if (debug&DEBUG_VERBOSE)
+	syslog(LOG_DEBUG,"running '%s'",buf);
+
+    if (running_pid != 0) unblock_signals();
+
+    if (running_pid < 0) {
+        syslog(LOG_ERR, "failed to fork and run '%s': %m",buf);
+	return -1;
+    }
+
+    if (running_pid == 0) {
+        /* change the signal actions back to the defaults, then unblock them. */
+        default_sigacts();
+	unblock_signals();
+
+        /* Leave the current location */
+        (void) setsid();    /* No controlling tty. */
+        (void) umask (S_IRWXG|S_IRWXO);
+        (void) chdir ("/"); /* no current directory. */
+
+	/* close all fd's the child should not see */
+	close(0);
+	close(1);
+	close(2);
+	close(modem_fd);
+	close(proxy_mfd);      /* close the master pty endpoint */
+	close(proxy_sfd);      /* close the slave pty endpoint */
+	if (fifo_fd != -1) close(fifo_fd);
+	if (monitors) {
+	    MONITORS *c = monitors;
+	    while (c) {
+		close(c->fd);
+		c = c->next;
+	    }
+	}
+
+	/* make sure the stdin, stdout and stderr get directed to /dev/null */
+	fd = open("/dev/null", O_RDWR);
+        if (fd >= 0) {
+	    if (fd != 0) {
+	    	dup2(fd, 0);
+		close(fd);
+	    }
+	    dup2(0, 1);
+            dup2(0, 2);
+        }
+
+        execl("/bin/sh", "sh", "-c", buf, (char *)0);
+        syslog(LOG_ERR, "could not exec /bin/sh: %m");
+        _exit(127);
+        /* NOTREACHED */
+    }
+    while (running_pid) {
+	pause();
+    }
+    return running_status;
+}
+
+void mon_write(char *message,int len)
+{
+    MONITORS *c = monitors, *p = 0, *cn;
+    while (c) {
+	cn = c->next;
+	if (write(c->fd,message,len) < 0) {
+	    if (errno == EPIPE) {
+	        syslog(LOG_INFO,"Monitor pipe closed.");
+	    } else {
+	        syslog(LOG_ERR,"Write error on monitor pipe. Closing pipe.");
+	    }
+	    close(c->fd);
+	    if (p) p->next = c->next;
+	    else monitors = c->next;
+	    free(c);
+	} else {
+	    p = c;
+	}
+	c = cn;
+    }
+}
