@@ -1,27 +1,24 @@
 /*
  * filter.c - Packet filtering for diald.
  *
- * Copyright (c) 1994, 1995 Eric Schenk.
+ * Copyright (c) 1994, 1995, 1996 Eric Schenk.
  * All rights reserved. Please see the file LICENSE which should be
  * distributed with this software for terms of use.
  */
 
 #include "diald.h"
 
+int itxtotal = 0;
+int irxtotal = 0;
+
 /*
- * Open up the various files we use to monitor network activity.
+ * Initialize the file descriptors for network monitoring sockets.
  */
 void filter_setup()
 {
-    /* get two socket to the INET kernel */ 
-    /* We need two sockets because some (recent) versions of the kernel
-     * won't forward packets back to the socket that sent the packet.
-     */
 
-    if ((fwdfd = socket(AF_INET, SOCK_PACKET, htons(ETH_P_ALL))) < 0) {
-        syslog(LOG_ERR, "Could not get socket to do packet forwarding: %m");
-        die(1);
-    }
+    fwdfd = -1;
+
     if ((snoopfd = socket(AF_INET, SOCK_PACKET, htons(ETH_P_ALL))) < 0) {
         syslog(LOG_ERR, "Could not get socket to do packet monitoring: %m");
         die(1);
@@ -33,6 +30,8 @@ void filter_setup()
  */
 void idle_filter_init()
 {
+    struct sockaddr to;
+
     if (mode == MODE_SLIP) {
 	sprintf(snoop_dev,"sl%d",link_iface);
     } else {
@@ -40,6 +39,25 @@ void idle_filter_init()
     }
     if (debug) syslog(LOG_INFO,"Changed snoop device to %s",snoop_dev);
     txtotal = rxtotal = 0;
+
+    /* try to bind the snooping socket to a particular device */
+    to.sa_family = AF_INET;
+    strcpy(to.sa_data,snoop_dev);
+    /* This bind may fail if the kernel isn't recent enough. */
+    /* This will just mean more work for the kernel. */
+    bind(snoopfd,&to,sizeof(struct sockaddr));
+
+    if (fwdfd != -1) {
+	close(fwdfd);
+	fwdfd = -1;
+    }
+    if ((fwdfd = socket(AF_INET, SOCK_PACKET, htons(ETH_P_ALL))) < 0) {
+        syslog(LOG_ERR, "Could not get socket to do packet forwarding: %m");
+        die(1);
+    }
+    /* This bind may fail if the kernel isn't recent enough. */
+    /* This will just mean more work for the kernel. */
+    bind(fwdfd,&to,sizeof(struct sockaddr));
 }
 
 /*
@@ -47,8 +65,22 @@ void idle_filter_init()
  */
 void idle_filter_proxy()
 {
+    struct sockaddr to;
+
+    if (fwdfd != -1) {
+	close(fwdfd);
+	fwdfd = -1;
+    }
     sprintf(snoop_dev,"sl%d",proxy_iface);
     if (debug) syslog(LOG_INFO,"Changed snoop device to %s",snoop_dev);
+
+    /* try to bind the snooping socket to a particular device */
+    /* Most likely this should close the old socket and open a new one first */
+    to.sa_family = AF_INET;
+    strcpy(to.sa_data,snoop_dev);
+    /* This bind may fail if the kernel isn't recent enough. */
+    /* This will just mean more work for the kernel. */
+    bind(snoopfd,&to,sizeof(struct sockaddr));
 }
 
 /*
@@ -63,20 +95,31 @@ void filter_read()
     int len;
 
     if ((len = recvfrom(snoopfd,packet,4096,0,&from,&from_len)) > 0) {
+	/* FIXME: really if the bind succeeds, then I don't need
+	 * this check. How can I shortcut this effectly?
+	 * perhaps two different filter_read routines?
+         */
 	if (strcmp(snoop_dev,from.sa_data) == 0) {
-#ifdef UNSAFE_ROUTING
-	    /* If we are doing unsafe routing, then we cannot count
-             * the transmitted packets on the forwarding side of the
-	     * transitter (since there is none!), so we attempt to
-	     * count them here. However, we can only tell packets
-	     * that are leaving our interface from this machine,
-	     * forwarded packets all get counted as received bytes.
-	     */
-	    if (((struct iphdr *)packet)->saddr == local_addr)
-		txtotal += len;
-	    else
-#endif
-	    rxtotal += len;
+	    if (do_reroute) {
+		/* If we are doing unsafe routing, then we cannot count
+		 * the transmitted packets on the forwarding side of the
+		 * transitter (since there is none!), so we attempt to
+		 * count them here. However, we can only tell packets
+		 * that are leaving our interface from this machine,
+		 * forwarded packets all get counted as received bytes.
+		 */
+		if (((struct iphdr *)packet)->saddr == local_addr) {
+		    txtotal += len;
+		    itxtotal += len;
+		} else {
+		    rxtotal += len;
+		    irxtotal += len;
+		}
+	    } else {
+	    	rxtotal += len;
+		irxtotal += len;
+	    }
+	
             if ((ntohs(((struct iphdr *)packet)->frag_off) & 0x1fff) == 0) {
 	        /* Mark passage of first packet */
 	        if (check_firewall(fwunit,packet,len) && state == STATE_UP)
@@ -126,7 +169,11 @@ void print_filter_queue(int sig)
 
 void monitor_queue()
 {
+    char buf[100];
     struct firewall_req req;
     req.unit = fwunit;
     ctl_firewall(IP_FW_MCONN,&req);
+    sprintf(buf,"LOAD\n%d\n%d\n",itxtotal,irxtotal);
+    itxtotal = irxtotal = 0;
+    mon_write(buf,strlen(buf));
 }
