@@ -1,22 +1,29 @@
 /*
  * ppp.c - ppp and pppd control.
  *
- * Copyright (c) 1994, 1995 Eric Schenk.
+ * Copyright (c) 1994, 1995, 1996 Eric Schenk.
  * All rights reserved. Please see the file LICENSE which should be
  * distributed with this software for terms of use.
  */
 
 #include "diald.h"
+
+#ifdef 0
 #ifdef PPP_VERSION_2_2_0
 #include <linux/ppp_defs.h>
 #include <linux/if_ppp.h>
 #else
 #include <linux/ppp.h>
 #endif
+#else
+#define PPPIOCGUNIT_2_1_2 0x5494
+#define PPPIOCGUNIT_2_2_0 _IOR('t', 86, int)
+#endif
 
 void ppp_start()
 {
     int pgrpid;
+    char buf[24];
 
     link_iface = -1;
     /* Run pppd directly here and set up to wait for the iface */
@@ -28,13 +35,16 @@ void ppp_start()
     }
 
     if (link_pid == 0) {
-	char **argv = (char **)malloc(sizeof(char *)*(pppd_argc+7));
+	char **argv = (char **)malloc(sizeof(char *)*(pppd_argc+9));
 	int i = 0, j;;
 
 	argv[i++] = PATH_PPPD;
 	argv[i++] = "-detach";
 	if (modem) argv[i++] = "modem";
 	if (crtscts) argv[i++] = "crtscts";
+        argv[i++] = "mtu";
+	sprintf(buf,"%d",mtu);
+	argv[i++] = buf;
 	if (netmask) {
 	    argv[i++] = "netmask";
 	    argv[i++] = netmask;
@@ -62,7 +72,7 @@ void ppp_start()
 	_exit(99);
 	/* NOTREACHED */
     }
-    syslog(LOG_ERR,"Running pppd (pid = %d).",link_pid);
+    syslog(LOG_INFO,"Running pppd (pid = %d).",link_pid);
 }
 
 /*
@@ -85,7 +95,13 @@ int ppp_set_addrs()
     ulong laddr = 0, raddr = 0;
 
     /* Try to get the interface number if we don't know it yet. */
-    if (link_iface == -1) ioctl(modem_fd, PPPIOCGUNIT, &link_iface);
+    if (link_iface == -1) {
+	 /* Try the pppd-2.2.0 ioctrl first,
+	  * Try the pppd-2.1.2 ioctrl if that fails
+	  */
+   	 if (ioctl(modem_fd, PPPIOCGUNIT_2_2_0, &link_iface) == -1)
+   	 	ioctl(modem_fd, PPPIOCGUNIT_2_1_2, &link_iface);
+    }
 
     /* Ok then, see if pppd has upped the interface yet. */
     if (link_iface != -1) {
@@ -113,25 +129,45 @@ int ppp_set_addrs()
 	else
 	   raddr = ((struct sockaddr_in *) &ifr.ifr_addr)->sin_addr.s_addr;
 
+ 	/* Check the MTU, see if it matches what we asked for. If it
+	 * doesn't warn the user and adjust the MTU setting.
+	 * (NOTE: Adjusting the MTU setting may cause kernel nastyness...)
+	 */
+	if (ioctl(snoopfd, SIOCGIFMTU, (caddr_t) &ifr) == -1) {
+	    syslog(LOG_ERR,"failed to get ppp mtu setting: %m");
+	} else {
+	    if (ifr.ifr_mtu != mtu) {
+	        syslog(LOG_ERR,"PPP negotiated mtu of %d does not match requested setting %d.",ifr.ifr_mtu,mtu);
+		syslog(LOG_ERR,"Attempting to auto adjust mtu.");
+		syslog(LOG_ERR,"Restart diald with mtu set to %d to avoid errors.",ifr.ifr_mtu);
+		mtu = ifr.ifr_mtu;
+	        proxy_config(local_ip,remote_ip);
+	    }
+	}
+
 	if (dynamic_addrs) {
 	    /* only do the configuration in dynamic mode. */
 	    struct in_addr addr;
 	    addr.s_addr = raddr;
-	    strcpy(remote_ip,inet_ntoa(addr));
+	    remote_ip = strdup(inet_ntoa(addr));
 	    addr.s_addr = laddr;
-	    strcpy(local_ip,inet_ntoa(addr));
+	    local_ip = strdup(inet_ntoa(addr));
 	    local_addr = laddr;
 	    syslog(LOG_INFO,"New addresses: local %s, remote %s.",
 		local_ip,remote_ip);
 	    proxy_config(local_ip,remote_ip);
-#ifndef UNSAFE_ROUTING
-	    add_routes("sl",proxy_iface,local_ip,remote_ip);
-#endif
+	    /* have to reset the proxy if we won't be rerouting... */
+	    if (!do_reroute)
+	        add_routes("sl",proxy_iface,local_ip,remote_ip,2000);
 	}
 
-#ifdef UNSAFE_ROUTING
-	add_routes("ppp",link_iface,local_ip,remote_ip);
-#endif
+	/* This is redundant in normal operation, but if we
+	 * have to restart the link, then this is necessary...
+	 */
+	set_ptp("ppp",link_iface,remote_ip,500);
+
+	if (do_reroute)
+	     add_routes("ppp",link_iface,local_ip,remote_ip,500);
 	return 1;
     }
     return 0;
@@ -144,13 +180,34 @@ int ppp_dead()
 	 * nothing to do with creating. Therefore, we have to reestablish
 	 * the proxy routes here, including the point to point route!
 	 */
-	set_ptp("sl",proxy_iface,orig_remote_ip);
-	add_routes("sl",proxy_iface,orig_local_ip,orig_remote_ip);
+	set_ptp("sl",proxy_iface,orig_remote_ip,2000);
+	add_routes("sl",proxy_iface,orig_local_ip,orig_remote_ip,2000);
     	local_addr = inet_addr(orig_local_ip);
     }
     return (link_pid == 0);
 }
 
+int ppp_route_exists()
+{
+    char buf[128];
+    int device = 0;
+    int found = 0;
+    FILE *fp;
+    sprintf(buf,"%s -n",PATH_ROUTE);
+    if ((fp = popen(buf,"r"))==NULL) {
+        syslog(LOG_ERR,"Could not run command '%s': %m",buf);
+	return 0;	/* assume half dead in this case... */
+    }
+
+    while (fgets(buf,128,fp)) {
+	if (sscanf(buf,"%*s %*s %*s %*s %*s %*s %*s ppp%d",&device) == 8) {
+	    if (device == link_iface) found = 1;
+	}
+    }
+    fclose(fp);
+    return found;
+}
+
 void ppp_stop()
 {
     if (link_pid)
@@ -162,14 +219,12 @@ void ppp_reroute()
 {
     /* Restore the original proxy routing */
     proxy_config(orig_local_ip,orig_remote_ip);
-    set_ptp("sl",proxy_iface,orig_remote_ip);
-    add_routes("sl",proxy_iface,orig_local_ip,orig_remote_ip);
+    set_ptp("sl",proxy_iface,orig_remote_ip,2000);
+    add_routes("sl",proxy_iface,orig_local_ip,orig_remote_ip,2000);
     local_addr = inet_addr(orig_local_ip);
-#ifdef UNSAFE_ROUTING
     /* If we did routing on the ppp link, remove it */
-    if (link_iface != -1)
-    	del_routes("ppp",link_iface,local_ip,remote_ip);
-#endif
+    if (do_reroute && link_iface != -1)
+    	del_routes("ppp",link_iface,local_ip,remote_ip,500);
     link_iface = -1;
 }
 
@@ -179,3 +234,11 @@ void ppp_kill()
     	if (kill(link_pid,SIGINT) == -1 && errno == ESRCH)
 	    link_pid = 0;
 }
+
+void ppp_zombie()
+{
+    /* Either ppp became a zombie or we missed a SIGCHLD signal */
+
+    sig_chld(SIGKILL);	/* try to reap the child */
+    link_pid = 0;	/* just in case the reaping failed, forget zombie */
+}
